from typing import Any, Dict

from unipipeline import UniWorker

from example.messages.ender_message import EnderMessage


class EnderSecondWorker(UniWorker[EnderMessage, EnderMessage]):
    def handle_message(self, message: EnderMessage) -> Dict[str, Any]:
        print(f'!!! message {message}')
        return {
            "some_prop": "WORLD"
        }



from time import sleep
from typing import Dict, TypeVar, Any, Set, Union, Optional, Type, List, NamedTuple, Tuple
from uuid import uuid4

from pydantic import ValidationError

from unipipeline.errors import UniPayloadError
from unipipeline.modules.uni_broker import UniBroker, UniBrokerConsumer
from unipipeline.modules.uni_config import UniConfig
from unipipeline.modules.uni_cron_job import UniCronJob
from unipipeline.modules.uni_echo import UniEcho
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_meta import UniMessageMeta
from unipipeline.modules.uni_worker import UniWorker
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition
from unipipeline.utils.sig import soft_interruption

TWorker = TypeVar('TWorker', bound=UniWorker)


class UniBrokerInitRecipe(NamedTuple):
    topics: Set[str]
    answer_topics: Set[str]


class UniMediator:
    def __init__(self, config: UniConfig) -> None:
        self._config = config

        self._worker_definition_by_type: Dict[Any, UniWorkerDefinition] = dict()
        self._worker_instance_indexes: Dict[str, UniWorker] = dict()
        self._broker_instance_indexes: Dict[str, UniBroker[Any]] = dict()
        self._worker_init_list: Set[str] = set()
        self._worker_initialized_list: Set[str] = set()
        self._waiting_init_list: Set[str] = set()
        self._waiting_initialized_list: Set[str] = set()

        self._consumers_list: Set[str] = set()
        self._brokers_with_topics_to_init: Dict[str, UniBrokerInitRecipe] = dict()
        self._brokers_with_topics_initialized: Dict[str, UniBrokerInitRecipe] = dict()

        self._message_types: Dict[str, Type[UniMessage]] = dict()

        self._brokers_with_active_consumption: List[UniBroker[Any]] = list()

    @property
    def echo(self) -> UniEcho:
        return self._config.echo

    def set_echo_level(self, level: int) -> None:
        self.echo.level = level

    def get_broker(self, name: str, singleton: bool = True) -> UniBroker:
        if not singleton:
            broker_def = self.config.brokers[name]
            broker_type = broker_def.type.import_class(UniBroker, self.echo)
            br = broker_type(mediator=self, definition=broker_def)
            return br
        if name not in self._broker_instance_indexes:
            self._broker_instance_indexes[name] = self.get_broker(name, singleton=False)
        return self._broker_instance_indexes[name]

    def add_worker_to_consume_list(self, name: str) -> None:
        wd = self._config.workers[name]
        if wd.marked_as_external:
            raise OverflowError(f'your could not use worker "{name}" as consumer. it marked as external "{wd.external}"')
        self._consumers_list.add(name)
        self.echo.log_info(f'added consumer {name}')

    def get_message_type(self, name: str) -> Type[UniMessage]:
        if name in self._message_types:
            return self._message_types[name]

        self._message_types[name] = self.config.messages[name].type.import_class(UniMessage, self.echo)

        return self._message_types[name]

    def answer_to(self, worker_name: str, meta: UniMessageMeta, payload: Optional[Union[Dict[str, Any], UniMessageMeta, UniMessage]]) -> None:
        wd = self._config.workers[worker_name]
        if not wd.need_answer:
            if payload is not None:
                raise UniPayloadError(f'output message must be None because worker {wd.name} has no possibility to send output messages')
            return

        if payload is None:
            raise UniPayloadError('output message must be not empty')

        answ_meta: UniMessageMeta
        if isinstance(payload, UniMessageMeta):
            answ_meta = payload
        else:
            assert wd.output_message is not None
            output_message_type = self.get_message_type(wd.output_message.name)
            payload_msg: UniMessage
            if isinstance(payload, output_message_type):
                payload_msg = payload
            elif isinstance(payload, dict):
                payload_msg = output_message_type(**payload)
            else:
                raise UniPayloadError(f'output message has invalid type. {type(payload).__name__} was given')

            answ_meta = meta.create_child(payload_msg.dict())

        b = self.get_broker(wd.broker.name)

        b.publish_answer(wd.answer_topic, meta.id, answ_meta)
        self.echo.log_info(f'worker {worker_name} answers to {wd.answer_topic}->{meta.id} :: {answ_meta}')

    def send_to(self, worker_name: str, payload: Union[Dict[str, Any], UniMessage], parent_meta: Optional[UniMessageMeta] = None, alone: bool = False) -> Optional[Tuple[UniMessage, UniMessageMeta]]:
        if worker_name not in self._worker_initialized_list:
            raise OverflowError(f'worker {worker_name} was not initialized')

        wd = self._config.workers[worker_name]

        message_type = self.get_message_type(wd.input_message.name)
        try:
            if isinstance(payload, message_type):
                payload_data = payload.dict()
            elif isinstance(payload, dict):
                payload_data = message_type(**payload).dict()
            else:
                raise TypeError(f'data has invalid type.{type(payload).__name__} was given')
        except ValidationError as e:
            raise UniPayloadError(e)

        br = self.get_broker(wd.broker.name)

        if alone:
            size = br.get_topic_approximate_messages_count(wd.topic)
            if size != 0:
                self.echo.log_info(f'sending to worker "{wd.name}" was skipped, because topic {wd.topic} has messages: {size}>0')
                return None

        if parent_meta is not None:
            meta = parent_meta.create_child(payload_data)
        else:
            meta = UniMessageMeta.create_new(payload_data)

        meta_list = [meta]
        br.publish(wd.topic, meta_list)  # TODO: make it list by default
        self.echo.log_info(f"worker {wd.name} sent message to topic '{wd.topic}':: {meta_list}")

        if wd.need_answer:
            assert wd.output_message is not None
            answ_meta = br.get_answer(wd.answer_topic, meta.id, 1)
            answ_message_type = self.get_message_type(wd.output_message.name)
            answ_msg = answ_message_type(**answ_meta.payload)
            return answ_msg, answ_meta
        return None

    def start_cron(self) -> None:
        cron_jobs = UniCronJob.mk_jobs_list(self.config.cron_tasks.values(), self)
        self.echo.log_debug(f'cron jobs defined: {", ".join(cj.task.name for cj in cron_jobs)}')
        while True:
            delay, jobs = UniCronJob.search_next_tasks(cron_jobs)
            if delay is None:
                return
            self.echo.log_debug(f"sleep {delay} seconds before running the tasks: {[cj.task.name for cj in jobs]}")
            if delay > 0:
                sleep(delay)
            self.echo.log_info(f"run the tasks: {[cj.task.name for cj in jobs]}")
            for cj in jobs:
                cj.send()
            sleep(1.1)  # delay for correct next iteration

    def start_consuming(self) -> None:
        brokers = set()
        for wn in self._consumers_list:
            wd = self._config.workers[wn]
            w = self.get_worker(wn)

            br = self.get_broker(wd.broker.name)

            self.echo.log_info(f"worker {wn} start consuming")
            br.add_consumer(UniBrokerConsumer(
                topic=wd.topic,
                id=f'{wn}__{uuid4()}',
                group_id=wn,
                message_handler=w.uni_process_message
            ))

            self.echo.log_info(f'consumer {wn} initialized')
            brokers.add(w.definition.broker.name)

        with soft_interruption(self._handle_interruption, self._handle_force_interruption, self._interruption_err):
            for bn in brokers:
                b = self.get_broker(bn)
                self._brokers_with_active_consumption.append(b)
                self.echo.log_info(f'broker {bn} consuming start')
                b.start_consuming()

    def _interruption_err(self, err: Exception) -> None:
        self.echo.log_error(str(err))
        raise err

    def _handle_force_interruption(self) -> None:
        self.echo.log_warning('force interruption detected')

    def _handle_interruption(self) -> None:
        self.echo.log_warning('interruption detected')
        for b in self._brokers_with_active_consumption:
            self.echo.log_debug(f'broker "{b.definition.name}" was notified about interruption')
            b.stop_consuming()
        self._brokers_with_active_consumption = list()
        self.echo.log_info(f'all brokers was notified about interruption')

    def add_worker_to_init_list(self, name: str, no_related: bool) -> None:
        if name not in self._config.workers:
            self.echo.exit_with_error(f'worker "{name}" is not found in config "{self.config.file}"')
        wd = self._config.workers[name]
        self._worker_init_list.add(name)
        for waiting in wd.waitings:
            if waiting.name not in self._waiting_initialized_list:
                self._waiting_init_list.add(waiting.name)
        self.add_broker_topic_to_init(wd.broker.name, wd.topic, False)
        self.add_broker_topic_to_init(wd.broker.name, wd.error_topic, False)
        self.add_broker_topic_to_init(wd.broker.name, wd.error_payload_topic, False)
        if wd.need_answer:
            self.add_broker_topic_to_init(wd.broker.name, wd.answer_topic, True)
        if not no_related:
            for wn in wd.output_workers:
                self._worker_init_list.add(wn)
                owd = self._config.workers[wn]
                self.add_broker_topic_to_init(owd.broker.name, owd.topic, False)

    def add_broker_topic_to_init(self, name: str, topic: str, is_answer: bool) -> None:
        if name in self._brokers_with_topics_initialized:
            if is_answer:
                if topic in self._brokers_with_topics_initialized[name].answer_topics:
                    return
            else:
                if topic in self._brokers_with_topics_initialized[name].topics:
                    return

        if name not in self._brokers_with_topics_to_init:
            self._brokers_with_topics_to_init[name] = UniBrokerInitRecipe(set(), set())

        if is_answer:
            self._brokers_with_topics_to_init[name].answer_topics.add(topic)
        else:
            self._brokers_with_topics_to_init[name].topics.add(topic)

    def initialize(self, create: bool = True) -> None:
        echo = self.echo.mk_child('initialize')
        for wn in self._worker_init_list:
            echo.log_info(f'worker "{wn}"', )
            self._worker_initialized_list.add(wn)
        self._worker_init_list = set()

        for waiting_name in self._waiting_init_list:
            self._config.waitings[waiting_name].wait(echo)
            echo.log_info(f'waiting "{waiting_name}"')
            self._waiting_initialized_list.add(waiting_name)
        self._waiting_init_list = set()

        if create:
            for bn, collection in self._brokers_with_topics_to_init.items():
                bd = self._config.brokers[bn]

                if bd.marked_as_external:
                    echo.log_debug(f'broker "{bn}" skipped because it external')
                    continue

                b = self.wait_for_broker_connection(bn)

                b.initialize(collection.topics, collection.answer_topics)
                echo.log_info(f'broker "{b.definition.name}" topics :: {collection.topics}')
                if len(collection.answer_topics) > 0:
                    echo.log_info(f'broker "{b.definition.name}" answer topics :: {collection.answer_topics}')

                if bn not in self._brokers_with_topics_initialized:
                    self._brokers_with_topics_initialized[bn] = UniBrokerInitRecipe(set(), set())
                for topic in collection.topics:
                    self._brokers_with_topics_initialized[bn].topics.add(topic)
                for topic in collection.answer_topics:
                    self._brokers_with_topics_initialized[bn].answer_topics.add(topic)
            self._brokers_with_topics_to_init = dict()

    def get_worker(self, worker: Union[Type['UniWorker'], str], singleton: bool = True) -> UniWorker:
        wd = self._config.get_worker_definition(worker)
        if wd.marked_as_external:
            raise OverflowError(f'worker "{worker}" is external. you could not get it')
        if not singleton or wd.name not in self._worker_instance_indexes:
            assert wd.type is not None
            worker_type = wd.type.import_class(UniWorker, self.echo)
            self.echo.log_info(f'get_worker :: initialized worker "{wd.name}"')
            w = worker_type(definition=wd, mediator=self)
        else:
            return self._worker_instance_indexes[wd.name]
        self._worker_instance_indexes[wd.name] = w
        return w

    @property
    def config(self) -> UniConfig:
        return self._config

    def wait_for_broker_connection(self, name: str) -> UniBroker:
        br = self.get_broker(name)
        for try_count in range(br.definition.retry_max_count):
            try:
                br.connect()
                self.echo.log_info(f'wait_for_broker_connection :: broker {br.definition.name} connected')
                return br
            except ConnectionError as e:
                self.echo.log_info(f'wait_for_broker_connection :: broker {br.definition.name} retry to connect [{try_count}/{br.definition.retry_max_count}] : {e}')
                sleep(br.definition.retry_delay_s)
                continue
        raise ConnectionError(f'unavailable connection to {br.definition.name}')



from pydantic import ValidationError


class UniError(Exception):
    pass


class UniDefinitionNotFoundError(UniError):
    pass


class UniConfigError(UniError):
    pass


class UniPayloadError(UniError):
    pass


class UniSendingToWorkerError(UniError):
    pass


class UniWorkFlowError(UniError):
    pass


class UniAnswerDelayError(UniError):
    pass



from typing import Generic, Type, Any, TypeVar, Optional, Dict, Union, Tuple

from unipipeline.errors import UniSendingToWorkerError
from unipipeline.modules.uni_broker import UniBrokerMessageManager
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_meta import UniMessageMeta, UniMessageMetaErrTopic
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition

TInputMessage = TypeVar('TInputMessage', bound=UniMessage)
TOutputMessage = TypeVar('TOutputMessage', bound=UniMessage)


class UniPayloadParsingError(Exception):
    def __init__(self, exception: Exception):
        self.parent_exception = exception


class UniWorker(Generic[TInputMessage, TOutputMessage]):
    def __init__(
        self,
        definition: UniWorkerDefinition,
        mediator: Any
    ) -> None:
        from unipipeline.modules.uni_mediator import UniMediator
        self._uni_moved = False
        self._uni_consume_initialized = False
        self._uni_payload_cache: Optional[TInputMessage] = None
        self._uni_current_meta: Optional[UniMessageMeta] = None
        self._uni_current_manager: Optional[UniBrokerMessageManager] = None
        self._uni_definition = definition
        self._uni_mediator: UniMediator = mediator
        self._uni_worker_instances_for_sending: Dict[Type[UniWorker], UniWorker] = dict()
        self._uni_echo = self._uni_mediator.echo.mk_child(f'worker[{self._uni_definition.name}]')
        self._uni_input_message_type: Type[TInputMessage] = self._uni_mediator.get_message_type(self._uni_definition.input_message.name)  # type: ignore
        self._uni_output_message_type: Optional[Type[TOutputMessage]] = self._uni_mediator.get_message_type(self._uni_definition.output_message.name) if self._uni_definition.output_message is not None else None  # type: ignore
        self._uni_echo_consumer = self._uni_echo.mk_child('consuming')
        self._uni_echo_consumer_sending = self._uni_echo_consumer.mk_child('sending')

    @property
    def input_message_type(self) -> Type[TInputMessage]:
        return self._uni_input_message_type

    @property
    def output_message_type(self) -> Optional[Type[TOutputMessage]]:
        return self._uni_output_message_type

    @property
    def definition(self) -> UniWorkerDefinition:
        return self._uni_definition

    @property
    def meta(self) -> UniMessageMeta:
        assert self._uni_current_meta is not None
        return self._uni_current_meta

    @property
    def manager(self) -> UniBrokerMessageManager:
        assert self._uni_current_manager is not None
        return self._uni_current_manager

    @property
    def payload(self) -> TInputMessage:
        return self._uni_payload_cache  # type: ignore

    def handle_message(self, message: TInputMessage) -> Optional[Union[TOutputMessage, Dict[str, Any]]]:
        raise NotImplementedError(f'method handle_message not implemented for {type(self).__name__}')

    def send_to(self, worker: Union[Type['UniWorker'], str], data: Any, alone: bool = False) -> Optional[Tuple[UniMessage, UniMessageMeta]]:
        if self._uni_current_meta is None:
            raise UniSendingToWorkerError(f'meta was not defined. incorrect usage of function "send_to"')
        wd = self._uni_mediator.config.get_worker_definition(worker)

        if wd.name not in self._uni_definition.output_workers:
            raise UniSendingToWorkerError(f'worker {wd.name} is not defined in workers->{self._uni_definition.name}->output_workers')

        return self._uni_mediator.send_to(wd.name, data, parent_meta=self._uni_current_meta, alone=alone)

    def uni_process_message(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        self._uni_echo_consumer.log_debug(f"message {meta.id} received :: {meta}")
        self._uni_reset_processing(meta, manager)

        err_topic = UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR
        try:
            self._uni_payload_cache = self._uni_input_message_type(**self.meta.payload)
            err_topic = UniMessageMetaErrTopic.HANDLE_MESSAGE_ERR
            result = self.handle_message(self._uni_payload_cache)
            self._uni_mediator.answer_to(self._uni_definition.name, meta, result)
        except Exception as e:
            self._uni_move_to_error_topic(err_topic, e)
        # else:
        #     try:
        #         assert meta.error is not None  # for mypy needs
        #         if meta.error.error_topic is UniMessageMetaErrTopic.HANDLE_MESSAGE_ERR:
        #             self.handle_error_message_handling(self.payload)
        #         elif meta.error.error_topic is UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR:
        #             self.handle_error_message_payload(self.meta, self.manager)
        #         elif meta.error.error_topic is UniMessageMetaErrTopic.ERROR_HANDLING_ERR:
        #             self.handle_error_handling(self.meta, self.manager)
        #         else:
        #             unsupported_err_topic = True
        #     except Exception as e:
        #         self._uni_echo_consumer.log_error(str(e))
        #         self.move_to_error_topic(UniMessageMetaErrTopic.ERROR_HANDLING_ERR, e)
        # if unsupported_err_topic:
        #     assert meta.error is not None  # for mypy needs
        #     err = NotImplementedError(f'{meta.error.error_topic} is not implemented in process_message')
        #     self._uni_echo_consumer.log_error(str(err))
        #     self.move_to_error_topic(UniMessageMetaErrTopic.SYSTEM_ERR, err)

        if not self._uni_moved and self._uni_definition.ack_after_success:
            manager.ack()

        self._uni_echo_consumer.log_info(f"message {meta.id} processed")
        self._uni_reset_processing()

    def _uni_reset_processing(self, meta: UniMessageMeta = None, manager: UniBrokerMessageManager = None) -> None:
        self._uni_moved = False
        self._uni_current_meta = meta
        self._uni_current_manager = manager
        self._uni_payload_cache = None

    def _uni_move_to_error_topic(self, err_topic: UniMessageMetaErrTopic, err: Exception) -> None:
        self._uni_echo_consumer.log_error(str(err))
        self._uni_moved = True
        meta = self.meta.create_error_child(err_topic, err)
        br = self._uni_mediator.get_broker(self._uni_definition.broker.name)
        error_topic = self.definition.error_topic
        if error_topic is UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR:
            error_topic = self.definition.error_payload_topic
        br.publish(error_topic, [meta])
        self.manager.ack()



import time
from datetime import timedelta
from time import sleep
from typing import Optional, TypeVar, Set, List, NamedTuple, Callable, Tuple
from urllib.parse import urlparse
from uuid import UUID

from pika import ConnectionParameters, PlainCredentials, BlockingConnection, BasicProperties, spec  # type: ignore
from pika.adapters.blocking_connection import BlockingChannel  # type: ignore
from pika.exceptions import AMQPConnectionError, AMQPError, ConnectionClosedByBroker  # type: ignore

from unipipeline.errors import UniAnswerDelayError
from unipipeline.modules.uni_broker import UniBroker, UniBrokerMessageManager, UniBrokerConsumer
from unipipeline.modules.uni_definition import UniDynamicDefinition
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_codec import UniMessageCodec
from unipipeline.modules.uni_message_meta import UniMessageMeta

BASIC_PROPERTIES__HEADER__COMPRESSION_KEY = 'compression'


TMessage = TypeVar('TMessage', bound=UniMessage)


class UniAmqpBrokerMessageManager(UniBrokerMessageManager):

    def __init__(self, channel: BlockingChannel, method_frame: spec.Basic.Deliver) -> None:
        self._channel = channel
        self._method_frame = method_frame
        self._acknowledged = False

    def reject(self) -> None:
        self._channel.basic_reject(delivery_tag=self._method_frame.delivery_tag, requeue=True)

    def ack(self) -> None:
        if self._acknowledged:
            return
        self._acknowledged = True
        self._channel.basic_ack(delivery_tag=self._method_frame.delivery_tag)


class UniAmqpBrokerConfig(UniDynamicDefinition):
    exchange_name: str = "communication"
    answer_exchange_name: str = "communication_answer"
    heartbeat: int = 600
    blocked_connection_timeout: int = 300
    prefetch: int = 1
    retry_max_count: int = 100
    retry_delay_s: int = 3
    socket_timeout: int = 300
    stack_timeout: int = 300
    exchange_type: str = "direct"
    durable: bool = True
    auto_delete: bool = False
    passive: bool = False
    is_persistent: bool = True


class UniAmqpBrokerConsumer(NamedTuple):
    queue: str
    on_message_callback: Callable[[BlockingChannel, spec.Basic.Deliver, BasicProperties, bytes], None]
    consumer_tag: str


class UniAmqpBroker(UniBroker[UniAmqpBrokerConfig]):
    config_type = UniAmqpBrokerConfig

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        ch = self._get_channel()
        res = ch.queue_declare(
            queue=topic,
            passive=True
        )
        return res.method.message_count

    @classmethod
    def get_connection_uri(cls) -> str:
        raise NotImplementedError(f"cls method get_connection_uri must be implemented for class '{cls.__name__}'")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        broker_url = self.get_connection_uri()

        url_params_pr = urlparse(url=broker_url)

        self._params = ConnectionParameters(
            heartbeat=self.config.heartbeat,
            blocked_connection_timeout=self.config.blocked_connection_timeout,
            socket_timeout=self.config.socket_timeout,
            stack_timeout=self.config.stack_timeout,
            retry_delay=self.definition.retry_delay_s,
            host=url_params_pr.hostname,
            port=url_params_pr.port,
            credentials=PlainCredentials(url_params_pr.username, url_params_pr.password, erase_on_connect=False),
        )

        self._consumers: List[UniAmqpBrokerConsumer] = list()

        self._connection: Optional[BlockingConnection] = None
        self._channel: Optional[BlockingChannel] = None

        self._consuming_started = False
        self._in_processing = False
        self._interrupted = False

    def initialize(self, topics: Set[str], answer_topic: Set[str]) -> None:
        ch = self._get_channel()
        ch.exchange_declare(
            exchange=self.config.exchange_name,
            exchange_type=self.config.exchange_type,
            passive=self.config.passive,
            durable=self.config.durable,
            auto_delete=self.config.auto_delete,
        )
        ch.exchange_declare(
            exchange=self.config.answer_exchange_name,
            exchange_type=self.config.exchange_type,
            passive=self.config.passive,
            durable=self.config.durable,
            auto_delete=self.config.auto_delete,
        )

        ch.basic_qos(prefetch_count=self.config.prefetch)

        for topic in topics:
            ch.queue_declare(
                queue=topic,
                durable=self.config.durable,
                auto_delete=self.config.auto_delete,
                passive=False,
            )

            ch.queue_bind(queue=topic, exchange=self.config.exchange_name, routing_key=topic)

    def stop_consuming(self) -> None:
        self._end_consuming()

    def _end_consuming(self):
        if not self._consuming_started:
            return
        self._interrupted = True
        if not self._in_processing:
            self._get_channel().stop_consuming()
            self.close()
            self._consuming_started = False
            self.echo.log_info('consumption stopped')

    def connect(self) -> None:
        if self._connection is not None:
            if self._connection.is_closed:
                self._connection = None
            else:
                return

        if self._channel is not None:
            if self._channel.is_closed:
                self._channel = None
            else:
                return

        try:
            self._connection = BlockingConnection(self._params)
            self._channel = self._connection.channel()
        except (AMQPError, AMQPConnectionError) as e:
            raise ConnectionError(str(e))

        self.echo.log_info('connected')

    def close(self) -> None:
        try:
            if self._channel is not None and not self._channel.is_closed:
                self._channel.close()
        except AMQPError:
            pass

        try:
            if self._connection is not None and not self._connection.is_closed:
                self._connection.close()
        except AMQPError:
            pass

        self._connection = None
        self._channel = None

    def _get_channel(self) -> BlockingChannel:
        self.connect()
        assert self._channel is not None
        return self._channel

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        echo = self.echo.mk_child(f'topic[{consumer.topic}]')
        if self._consuming_started:
            echo.log_error(f'you cannot add consumer dynamically :: tag="{consumer.id}" group_id={consumer.group_id}')
            exit(1)

        def consumer_wrapper(channel: BlockingChannel, method_frame: spec.Basic.Deliver, properties: BasicProperties, body: bytes) -> None:
            self._in_processing = True
            meta = self.codec_parse(body, UniMessageCodec(
                compression=properties.headers.get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
                content_type=properties.content_type
            ))
            manager = UniAmqpBrokerMessageManager(channel, method_frame)
            consumer.message_handler(meta, manager)
            self._in_processing = False
            if self._interrupted:
                self._end_consuming()

        self._consumers.append(UniAmqpBrokerConsumer(
            queue=consumer.topic,
            on_message_callback=consumer_wrapper,
            consumer_tag=consumer.id,
        ))

        echo.log_info(f'added consumer :: tag="{consumer.id}" group_id={consumer.group_id}')

    def start_consuming(self) -> None:
        echo = self.echo.mk_child('consuming')
        if len(self._consumers) == 0:
            echo.log_warning('has no consumers to start consuming')
            return
        if self._consuming_started:
            echo.log_warning('consuming has already started. ignored')
            return
        self._consuming_started = True
        self._interrupted = False
        self._in_processing = False

        retry_counter = 0
        retry_threshold_s = self.config.retry_delay_s * (self.config.retry_max_count + 1)
        while True:
            start = time.time()
            try:
                ch = self._get_channel()
                for c in self._consumers:
                    ch.basic_consume(queue=c.queue, on_message_callback=c.on_message_callback, consumer_tag=c.consumer_tag)
                echo.log_info(f'consumers count is {len(self._consumers)}')
                ch.start_consuming()  # blocking operation
            except (ConnectionClosedByBroker, ConnectionError) as e:
                end = time.time()
                echo.log_error(f'connection closed {e}')
                if int(end - start) >= retry_threshold_s:
                    retry_counter = 0
                if retry_counter >= self.config.retry_max_count:
                    raise ConnectionError()
                retry_counter += 1
                sleep(self.config.retry_delay_s)

    def publish(self, topic: str, meta_list: List[UniMessageMeta]) -> None:
        for meta in meta_list: # TODO: package sending
            # TODO: retry
            self._get_channel().basic_publish(
                exchange=self.config.exchange_name,
                routing_key=topic,
                body=self.codec_serialize(meta),
                properties=BasicProperties(
                    content_type=self.definition.codec.content_type,
                    content_encoding='utf-8',
                    delivery_mode=2 if self.config.is_persistent else 0,
                    headers={
                        BASIC_PROPERTIES__HEADER__COMPRESSION_KEY: self.definition.codec.compression
                    }
                )
            )

    def get_answer(self, answer_topic: str, answer_id: UUID, max_delay_s: int) -> UniMessageMeta:
        topic = f'{answer_topic}.{answer_id}' if answer_topic else str(answer_id)
        ch = self._get_channel()

        ch.queue_declare(
            queue=topic,
            durable=False,
            auto_delete=True,
            passive=False,
        )

        started = time.time()
        while True:
            (method, properties, body) = ch.basic_get(topic, auto_ack=True)

            if method is None:
                if (time.time() - started) > max_delay_s:
                    raise UniAnswerDelayError(f'answer for {topic} reached limit')
                sleep(1)
                continue

            return self.codec_parse(body, UniMessageCodec(
                compression=properties.headers.get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
                content_type=properties.content_type
            ))

    def publish_answer(self, answer_topic: str, answer_id: UUID, meta: UniMessageMeta) -> None:
        ch = self._get_channel()
        topic = f'{answer_topic}.{answer_id}' if answer_topic else str(answer_id)
        ch.queue_declare(
            queue=topic,
            durable=False,
            auto_delete=True,
            passive=False,
        )
        ch.basic_publish(
            exchange=self.config.answer_exchange_name,
            routing_key=topic,
            body=self.codec_serialize(meta),
            properties=BasicProperties(
                content_type=self.definition.codec.content_type,
                content_encoding='utf-8',
                delivery_mode=0,
                headers={
                    BASIC_PROPERTIES__HEADER__COMPRESSION_KEY: self.definition.codec.compression
                }
            )
        )
